@@ -9,19 +9,27 @@ import os
 import shlex
 import typing as t
 
+from libtmux.common import tmux_cmd
+from libtmux.pane import Pane
+
 from . import exc, formats
 from .common import (
     PaneDict,
     TmuxMappingObject,
     TmuxRelationalObject,
+    WindowDict,
+    WindowOptionDict,
     handle_option_error,
 )
-from .pane import Pane
+
+if t.TYPE_CHECKING:
+    from .server import Server
+    from .session import Session
 
 logger = logging.getLogger(__name__)
 
 
-class Window(TmuxMappingObject, TmuxRelationalObject):
+class Window(TmuxMappingObject, TmuxRelationalObject["Pane", "PaneDict"]):
     """
     A :term:`tmux(1)` :term:`Window` [window_manual]_.
 
@@ -31,6 +39,32 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
     ----------
     session : :class:`Session`
 
+    Examples
+    --------
+    >>> window = session.new_window('My project')
+
+    >>> window
+    Window(@... ...:My project, Session($... ...))
+
+    Windows have panes:
+
+    >>> window.panes
+    [Pane(...)]
+
+    >>> window.attached_pane
+    Pane(...)
+
+    Relations moving up:
+
+    >>> window.session
+    Session(...)
+
+    >>> window == session.attached_window
+    True
+
+    >>> window in session.windows
+    True
+
     References
     ----------
     .. [window_manual] tmux window. openbsd manpage for TMUX(1).
@@ -41,25 +75,24 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
        https://man.openbsd.org/tmux.1#DESCRIPTION. Accessed April 1st, 2018.
     """
 
-    #: unique child ID key for :class:`~libtmux.common.TmuxRelationalObject`
     child_id_attribute = "pane_id"
-    #: namespace used :class:`~libtmux.common.TmuxMappingObject`
+    """Unique child ID key for :class:`~libtmux.common.TmuxRelationalObject`"""
     formatter_prefix = "window_"
+    """Namespace used for :class:`~libtmux.common.TmuxMappingObject`"""
+    server: "Server"
+    """:class:`libtmux.Server` window is linked to"""
+    session: "Session"
+    """:class:`libtmux.Session` window is linked to"""
 
-    def __init__(self, session=None, **kwargs):
-
-        if not session:
-            raise ValueError("Window requires a Session, session=Session")
-
+    def __init__(
+        self, session: "Session", window_id: t.Union[int, str], **kwargs: t.Any
+    ) -> None:
         self.session = session
         self.server = self.session.server
 
-        if "window_id" not in kwargs:
-            raise ValueError("Window requires a `window_id`")
+        self._window_id = window_id
 
-        self._window_id = kwargs["window_id"]
-
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "{}({} {}:{}, {})".format(
             self.__class__.__name__,
             self.id,
@@ -69,12 +102,11 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         )
 
     @property
-    def _info(self, *args):
-
+    def _info(self) -> WindowDict:  # type: ignore  # mypy#1362
         attrs = {"window_id": self._window_id}
 
         # from https://github.com/serkanyersen/underscore.py
-        def by(val) -> bool:
+        def by(val: WindowDict) -> bool:
             for key in attrs.keys():
                 try:
                     if attrs[key] != val[key]:
@@ -83,8 +115,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
                     return False
             return True
 
-        # TODO add type hint
-        target_windows = list(filter(by, self.server._windows))
+        target_windows = [s for s in self.server._windows if by(s)]
         # If a window_shell option was configured which results in
         # a short-lived process, the window id is @0.  Use that instead of
         # self._window_id
@@ -92,7 +123,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
             target_windows = self.server._windows
         return target_windows[0]
 
-    def cmd(self, cmd, *args, **kwargs):
+    def cmd(self, cmd: str, *args: t.Any, **kwargs: t.Any) -> tmux_cmd:
         """Return :meth:`Server.cmd` defaulting ``target_window`` as target.
 
         Send command to tmux with :attr:`window_id` as ``target-window``.
@@ -115,7 +146,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
 
         return self.server.cmd(cmd, *args, **kwargs)
 
-    def select_layout(self, layout=None):
+    def select_layout(self, layout: t.Optional[str] = None) -> None:
         """Wrapper for ``$ tmux select-layout <layout>``.
 
         Parameters
@@ -154,7 +185,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         if proc.stderr:
             raise exc.LibTmuxException(proc.stderr)
 
-    def set_window_option(self, option, value):
+    def set_window_option(self, option: str, value: t.Union[int, str]) -> None:
         """
         Wrapper for ``$ tmux set-window-option <option> <value>``.
 
@@ -190,17 +221,19 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         if isinstance(cmd.stderr, list) and len(cmd.stderr):
             handle_option_error(cmd.stderr[0])
 
-    def show_window_options(self, option=None, g=False):
+    def show_window_options(self, g: t.Optional[bool] = False) -> WindowOptionDict:
         """
         Return a dict of options for the window.
 
         For familiarity with tmux, the option ``option`` param forwards to
         pick a single option, forwarding to :meth:`Window.show_window_option`.
 
+        .. versionchanged:: 0.13.0
+
+           ``option`` removed, use show_window_option to return an individual option.
+
         Parameters
         ----------
-        option : str, optional
-            show a single option.
         g : str, optional
             Pass ``-g`` flag for global variable, default False.
 
@@ -208,32 +241,34 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         -------
         dict
         """
-
-        tmux_args = tuple()
+        tmux_args: t.Tuple[str, ...] = tuple()
 
         if g:
             tmux_args += ("-g",)
 
-        if option:
-            return self.show_window_option(option, g=g)
-        else:
-            tmux_args += ("show-window-options",)
-            cmd = self.cmd(*tmux_args).stdout
+        tmux_args += ("show-window-options",)
+        cmd = self.cmd(*tmux_args)
+
+        output = cmd.stdout
 
         # The shlex.split function splits the args at spaces, while also
         # retaining quoted sub-strings.
         #   shlex.split('this is "a test"') => ['this', 'is', 'a test']
-        cmd = [tuple(shlex.split(item)) for item in cmd]
 
-        window_options = dict(cmd)
+        window_options: WindowOptionDict = {}
+        for item in output:
+            key, val = shlex.split(item)
+            assert isinstance(key, str)
+            assert isinstance(val, str)
 
-        for key, value in window_options.items():
-            if value.isdigit():
-                window_options[key] = int(value)
+            if isinstance(val, str) and val.isdigit():
+                window_options[key] = int(val)
 
         return window_options
 
-    def show_window_option(self, option, g=False):
+    def show_window_option(
+        self, option: str, g: bool = False
+    ) -> t.Optional[t.Union[str, int]]:
         """
         Return a list of options for the window.
 
@@ -254,8 +289,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         :exc:`exc.OptionError`, :exc:`exc.UnknownOption`,
         :exc:`exc.InvalidOption`, :exc:`exc.AmbiguousOption`
         """
-
-        tmux_args = tuple()
+        tmux_args: t.Tuple[t.Union[str, int], ...] = tuple()
 
         if g:
             tmux_args += ("-g",)
@@ -267,15 +301,18 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         if len(cmd.stderr):
             handle_option_error(cmd.stderr[0])
 
-        if not len(cmd.stdout):
+        window_options_output = cmd.stdout
+
+        if not len(window_options_output):
             return None
 
-        option = [shlex.split(item) for item in cmd.stdout][0]
+        value_raw = [shlex.split(item) for item in window_options_output][0]
 
-        if option[1].isdigit():
-            option = (option[0], int(option[1]))
+        value: t.Union[str, int] = (
+            int(value_raw[1]) if value_raw[1].isdigit() else value_raw[1]
+        )
 
-        return option[1]
+        return value
 
     def rename_window(self, new_name: str) -> "Window":
         """
@@ -285,6 +322,17 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         ----------
         new_name : str
             name of the window
+
+        Examples
+        --------
+
+        >>> window = session.attached_window
+
+        >>> window.rename_window('My project')
+        Window(@1 ...:My project, Session($1 ...))
+
+        >>> window.rename_window('New name')
+        Window(@1 ...:New name, Session($1 ...))
         """
 
         import shlex
@@ -303,7 +351,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
 
         return self
 
-    def kill_window(self):
+    def kill_window(self) -> None:
         """Kill the current :class:`Window` object. ``$ tmux kill-window``."""
 
         proc = self.cmd(
@@ -317,7 +365,9 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
 
         self.server._update_windows()
 
-    def move_window(self, destination="", session=None):
+    def move_window(
+        self, destination: str = "", session: t.Optional[str] = None
+    ) -> None:
         """
         Move the current :class:`Window` object ``$ tmux move-window``.
 
@@ -356,7 +406,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         """
         return self.session.select_window(self.index)
 
-    def select_pane(self, target_pane: str) -> t.Optional[Pane]:
+    def select_pane(self, target_pane: t.Union[str, int]) -> t.Optional[Pane]:
         """
         Return selected :class:`Pane` through ``$ tmux select-pane``.
 
@@ -386,12 +436,12 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
 
     def split_window(
         self,
-        target=None,
-        start_directory=None,
-        attach=True,
-        vertical=True,
-        shell=None,
-        percent=None,
+        target: t.Optional[t.Union[int, str]] = None,
+        start_directory: t.Optional[str] = None,
+        attach: bool = True,
+        vertical: bool = True,
+        shell: t.Optional[str] = None,
+        percent: t.Optional[int] = None,
     ) -> Pane:
         """
         Split window and return the created :class:`Pane`.
@@ -445,7 +495,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
 
         # '-t%s' % self.attached_pane.get('pane_id'),
         # 2013-10-18 LOOK AT THIS, rm'd it..
-        tmux_args = tuple()
+        tmux_args: t.Tuple[str, ...] = tuple()
 
         if target:
             tmux_args += ("-t%s" % target,)
@@ -473,24 +523,25 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         if shell:
             tmux_args += (shell,)
 
-        pane = self.cmd("split-window", *tmux_args)
+        pane_cmd = self.cmd("split-window", *tmux_args)
 
         # tmux < 1.7. This is added in 1.7.
-        if pane.stderr:
-            raise exc.LibTmuxException(pane.stderr)
-            if "pane too small" in pane.stderr:
-                pass
+        if pane_cmd.stderr:
+            if "pane too small" in pane_cmd.stderr:
+                raise exc.LibTmuxException(pane_cmd.stderr)
 
-            raise exc.LibTmuxException(pane.stderr, self._info, self.panes)
-        else:
-            pane = pane.stdout[0]
+            raise exc.LibTmuxException(pane_cmd.stderr, self._info, self.panes)
 
-            pane = dict(zip(pformats, pane.split(formats.FORMAT_SEPARATOR)))
+        pane_output = pane_cmd.stdout[0]
 
-            # clear up empty dict
-            pane = {k: v for k, v in pane.items() if v}
+        pane_formatters = dict(
+            zip(pformats, pane_output.split(formats.FORMAT_SEPARATOR))
+        )
 
-        return Pane(window=self, **pane)
+        # Prune empty values
+        pane_formatters_filtered = {k: v for k, v in pane_formatters.items() if v}
+
+        return Pane(window=self, **pane_formatters_filtered)
 
     @property
     def attached_pane(self) -> t.Optional[Pane]:
@@ -505,6 +556,7 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
             # for now pane_active is a unicode
             if "pane_active" in pane and pane.get("pane_active") == "1":
                 return Pane(window=self, **pane)
+        return None
 
     def _list_panes(self) -> t.List[PaneDict]:
         panes = self.server._update_panes()._panes
@@ -536,4 +588,4 @@ class Window(TmuxMappingObject, TmuxRelationalObject):
         return self.list_panes()
 
     #: Alias :attr:`panes` for :class:`~libtmux.common.TmuxRelationalObject`
-    children = panes
+    children = panes  # type:ignore  # mypy#1362
